@import url('https://fonts.googleapis.com/css2?family=Lato:wght@300&display=swap');

h1, h2, h3 {
    font-family: 'Lato', sans-serif;
}

h1, h2 {
    color: #f47228;
}

html[data-theme=dark] .bd-content img:not(.only-dark,.dark-light) {

 background-color: rgba(0, 0, 0, 0.0);
}